
package ch.ethz.ssh2;

import java.io.IOException;

import ch.ethz.ssh2.auth.AuthenticationManager;
import ch.ethz.ssh2.channel.ChannelManager;
import ch.ethz.ssh2.crypto.CryptoWishList;
import ch.ethz.ssh2.crypto.cipher.BlockCipherFactory;
import ch.ethz.ssh2.crypto.digest.MAC;
import ch.ethz.ssh2.transport.TransportManager;

/**
 * A <code>Connection</code> is used to establish an encrypted TCP/IP
 * connection to a SSH2 server.<br>
 * <br>
 * Typically, one
 * <ul>
 * <li>Creates a <code>Connection</code> object.</li>
 * <li>Calls the <code>connect</code> method.</li>
 * <li>Calls some of the authentication methods.</li>
 * <li>Calls one or several times the <code>openSession</code>,
 * <code>createPortForwarder</code> or similar methods.</li>
 * <li>Finally, one can close the connection with the <code>close</code>
 * method.</li>
 * </ul>
 * 
 * @author Christian Plattner, plattner@inf.ethz.ch
 * @version $Id: Connection.java,v 1.9 2005/07/21 15:17:48 cplattne Exp $
 */

public class Connection
{
	private AuthenticationManager am;
	private boolean authenticated = false;

	private ChannelManager cm;

	private CryptoWishList cryptoWishList = new CryptoWishList();
	private DHGexParameters dhgexpara = new DHGexParameters();

	private String hostname;
	private int port;

	private TransportManager tm;

	public final static String identification = "Ganymed_build207";

	/**
	 * Same as <code>Connection(hostname, 22)</code<. 
	 * 
	 * @param hostname the hostname of the SSH2 server.
	 */
	public Connection(String hostname)
	{
		this(hostname, 22);
	}

	/**
	 * Prepares a fresh <code>Connection</code> object which can then be used
	 * to establish a connection to the specified SSH2 server.
	 * 
	 * @param hostname
	 *            the host where we later want to connect.
	 * @param port
	 *            port on the server, normally 22.
	 */
	public Connection(String hostname, int port)
	{
		this.hostname = hostname;
		this.port = port;
	}

	/**
	 * After a successful connect, one has to authenticate oneself. This method
	 * can be used to tell which authentication methods are supported by the
	 * server at a certain stage of the authentication process (for the given
	 * username).<br>
	 * Note 1: the username will only be used if no authentication step was done
	 * so far (it will be used to ask the server for a list of possible
	 * authentication methods). Otherwise, this method ignores the user name and
	 * returns a cached method list (which is based on the information contained
	 * in the last negative server response). <br>
	 * Note 2: the server may return method names that are not supported by this
	 * implementation.<br>
	 * After a successful authentication, this method must not be called
	 * anymore.
	 * 
	 * @param user
	 *            A <code>String</code> holding the username.
	 * 
	 * @return a (possibly emtpy) array holding authentication method names.
	 * @throws IOException
	 */
	public synchronized String[] getRemainingAuthMethods(String user) throws IOException
	{
		if (user == null)
			throw new IllegalArgumentException("user may not be NULL!");

		if (tm == null)
			throw new IllegalStateException("Connection is not established!");

		if (authenticated)
			throw new IllegalStateException("Connection is already authenticated!");

		if (am == null)
			am = new AuthenticationManager(tm);

		if (cm == null)
			cm = new ChannelManager(tm);

		return am.getRemainingMethods(user);
	}

	/**
	 * Checks if a specified authentication method is available. This method is
	 * actually just a wrapper for <code>getRemainingAuthMethods</code>.
	 * 
	 * @param user
	 *            A <code>String</code> holding the username.
	 * @param method
	 *            An authentication method name (e.g., "publickey", "password",
	 *            "keyboard-interactive") as specified by the SSH-2 standard.
	 * @return if the specified authentication method is currently available.
	 * @throws IOException
	 */
	public synchronized boolean isAuthMethodAvailable(String user, String method) throws IOException
	{
		if (method == null)
			throw new IllegalArgumentException("method may not be NULL!");

		String methods[] = getRemainingAuthMethods(user);

		for (int i = 0; i < methods.length; i++)
		{
			if (methods[i].compareTo(method) == 0)
				return true;
		}

		return false;
	}

	/**
	 * Returns true if there was at least on failed authentication request and
	 * the last failed authentication request was marked with "partial success"
	 * by the server. This is only needed in the rare case of SSH2 server setups
	 * that cannot be satisfied with a single successful authentication request
	 * (i.e., multiple authentication steps are needed.)
	 * 
	 * If you are interested in the details, then have a look at
	 * draft-ietf-secsh-userauth-XX.txt.
	 * 
	 * @return if the there was a failed authentication step and the last one
	 *         was marked as a "partial success".
	 */
	public synchronized boolean isAuthenticationPartialSuccess()
	{
		if (am == null)
			return false;

		return am.getPartialSuccess();
	}

	/**
	 * Determines if the authentication phase is complete. Can be called at any
	 * time.
	 * 
	 * @return <code>true</code> if no further authentication steps are
	 *         needed.
	 */
	public synchronized boolean isAuthenticationComplete()
	{
		return authenticated;
	}

	/**
	 * After a successful connect, one has to authenticate oneself. This method
	 * is based on DSA (it uses DSA to sign a challenge sent by the server). At
	 * the moment, only unencrypted and 3DES encrypted DSA private keys are
	 * supported. TODO: add support for DES, AES-128, AES-192 and AES-256
	 * encrypted private keys. <br>
	 * If the authentication phase is complete, <code>true</code> will be
	 * returned. If the server does not accept the request (or if further
	 * authentication steps are needed), <code>false</code> is returned and
	 * one can retry either by using this or any other authentication method
	 * (use the <code>getRemainingAuthMethods</code> method to get a list of
	 * the remaining possible methods).
	 * 
	 * @param user
	 *            A <code>String</code> holding the username.
	 * @param pem
	 *            A <code>String</code> containing the DSA private key of the
	 *            user in OpenSSH key format (PEM, you can't miss the
	 *            "-----BEGIN DSA PRIVATE KEY-----" tag). The string may contain
	 *            linefeeds.
	 * @param password
	 *            If the PEM string is 3DES encrypted ("DES-EDE3-CBC"), then you
	 *            must specify the password. Otherwise, this argument will be
	 *            ignored and can be set to <code>null</code>.
	 * 
	 * @return whether the connection is now authenticated.
	 * @throws IOException
	 */
	public synchronized boolean authenticateWithDSA(String user, String pem, String password) throws IOException
	{
		if (tm == null)
			throw new IllegalStateException("Connection is not established!");

		if (authenticated)
			throw new IllegalStateException("Connection is already authenticated!");

		if (am == null)
			am = new AuthenticationManager(tm);

		if (cm == null)
			cm = new ChannelManager(tm);

		authenticated = am.authenticateDSA(user, pem, password);

		return authenticated;
	}

	/**
	 * A wrapper that calls <code>authenticateWithKeyboardInteractive()</code>
	 * with a <code>null</code> submethod list.
	 * 
	 * @param user
	 *            A <code>String</code> holding the username.
	 * @param cb
	 *            An <code>InteractiveCallback</code> which will be used to
	 *            determine the responses to the questions asked by the server.
	 * @return whether the connection is now authenticated.
	 * @throws IOException
	 */
	public synchronized boolean authenticateWithKeyboardInteractive(String user, InteractiveCallback cb)
			throws IOException
	{
		return authenticateWithKeyboardInteractive(user, null, cb);
	}

	/**
	 * After a successful connect, one has to authenticate oneself. This method
	 * is based on "keyboard-interactive", specified in
	 * draft-ietf-secsh-auth-kbdinteract-XX. Basically, you have to define a
	 * callback object which will be feeded with challenges generated by the
	 * server. Answers are then sent back to the server. It is possible that the
	 * callback will be called several times during the invocation of this
	 * method (e.g., if the server replies to the callback's answer(s) with
	 * another challenge...) <br>
	 * If the authentication phase is complete, <code>true</code> will be
	 * returned. If the server does not accept the request (or if further
	 * authentication steps are needed), <code>false</code> is returned and
	 * one can retry either by using this or any other authentication method
	 * (use the <code>getRemainingAuthMethods</code> method to get a list of
	 * the remaining possible methods).<br>
	 * Note: some SSH servers advertise "keyboard-interactive", however, any
	 * interactive request will be denied (without having sent any challenge to
	 * the client).
	 * 
	 * @param user
	 *            A <code>String</code> holding the username.
	 * @param submethods
	 *            An array of submethod names, see
	 *            draft-ietf-secsh-auth-kbdinteract-XX. May be <code>null</code>
	 *            to indicate an empty list.
	 * @param cb
	 *            An <code>InteractiveCallback</code> which will be used to
	 *            determine the responses to the questions asked by the server.
	 * 
	 * @return whether the connection is now authenticated.
	 * @throws IOException
	 */
	public synchronized boolean authenticateWithKeyboardInteractive(String user, String[] submethods,
			InteractiveCallback cb) throws IOException
	{
		if (cb == null)
			throw new IllegalArgumentException("Callback may not ne NULL!");

		if (tm == null)
			throw new IllegalStateException("Connection is not established!");

		if (authenticated)
			throw new IllegalStateException("Connection is already authenticated!");

		if (am == null)
			am = new AuthenticationManager(tm);

		if (cm == null)
			cm = new ChannelManager(tm);

		authenticated = am.authenticateInteractive(user, submethods, cb);

		return authenticated;
	}

	/**
	 * After a successfull connect, one has to authenticate oneself. This method
	 * sends username and password to the server.<br>
	 * If the authentication phase is complete, <code>true</code> will be
	 * returned. If the server does not accept the request (or if further
	 * authentication steps are needed), <code>false</code> is returned and
	 * one can retry either by using this or any other authentication method
	 * (use the <code>getRemainingAuthMethods</code> method to get a list of
	 * the remaining possible methods).
	 * 
	 * @param user
	 * @param pass
	 * @return if the connection is now authenticated.
	 * @throws IOException
	 */
	public synchronized boolean authenticateWithPassword(String user, String pass) throws IOException
	{
		if (tm == null)
			throw new IllegalStateException("Connection is not established!");

		if (authenticated)
			throw new IllegalStateException("Connection is already authenticated!");

		if (am == null)
			am = new AuthenticationManager(tm);

		if (cm == null)
			cm = new ChannelManager(tm);

		authenticated = am.authenticatePassword(user, pass);

		return authenticated;
	}

	/**
	 * Close the connection to the SSH2 server. All assigned sessions will be
	 * closed, too. Can be called at any time. Don't forget to call this once
	 * you don't need a connection anymore - otherwise the receiver thread will
	 * run forever.
	 */
	public synchronized void close()
	{
		if (tm != null)
		{
			tm.close("Closed due to user request.");
			tm = null;
		}
		am = null;
		cm = null;
		authenticated = false;
	}

	/**
	 * Same as <code>connect(null)</code>.
	 * 
	 * @return see comments for the <code>connect(null)</code> method.
	 * @throws IOException
	 */
	public synchronized ConnectionInfo connect() throws IOException
	{
		return connect(null);
	}

	/**
	 * Connect to the SSH2 server and, as soon as the server has presented its
	 * host key, use the <code>verifyServerHostKey</code> method of the
	 * <code>verifier</code> to ask for permission to proceed. Only correctly
	 * signed host keys will be passed to the <code>verifier</code>. If
	 * <code>verifier</code> is <code>null</code>, then any correctly
	 * signed host key will be accepted - this is NOT recommended, since it
	 * makes man-in-the-middle attackes VERY easy.<br>
	 * <br>
	 * Note: as long as the key exchange of the underlying connection has not
	 * been completed, this call will block.
	 * 
	 * @param verifier
	 *            An object that implements the
	 *            <code>ServerHostKeyVerifier</code> interface.
	 * @return A <code>ConnectionInfo</code> object containg the details of
	 *         the established connection.
	 * @throws IOException
	 *             If any problem occurs, e.g., the server's host key is not
	 *             correctly signed or not accepted by the <code>verifier</code>.
	 */
	public synchronized ConnectionInfo connect(ServerHostKeyVerifier verifier) throws IOException
	{
		if (tm != null)
			throw new IOException("Connection is already in connected state!");

		try
		{
			tm = new TransportManager(hostname, port);
			tm.initialize(cryptoWishList, verifier, dhgexpara);

			/* Wait until first KEX has finished */

			return tm.getConnectionInfo(1);
		}
		catch (IOException e1)
		{
			close();
			throw new IOException(e1.getMessage());
		}
	}

	/**
	 * Force an asynchronous key re-exchange (the call does not block). The
	 * latest values set for MAC, Cipher and DH group exchange parameters will
	 * be used. If a key exchange is currently in progress, then this method has
	 * the only effect that the so far specified parameters will be used for the
	 * next (server driven) key exchange. NOTE: This implementation will never
	 * start a key exchange (other than the initial one) unless you or the SSH2
	 * server ask for it.
	 * 
	 * @throws IOException
	 *             In case of any failure behind the scenes.
	 */
	public synchronized void forceKeyExchange() throws IOException
	{
		if (tm == null)
			throw new IllegalStateException("You need to establish a connection first.");

		tm.forceKeyExchange(cryptoWishList, dhgexpara);
	}

	/**
	 * Unless you know what you are doing, you will never need this.
	 * 
	 * @return The list of supported cipher algorithms by this implementation.
	 */
	public synchronized String[] getAvailableCiphers()
	{
		return BlockCipherFactory.getCipherList();
	}

	/**
	 * Unless you know what you are doing, you will never need this.
	 * 
	 * @return The list of supported MAC algorthims by this implementation.
	 */
	public synchronized String[] getAvailableMACs()
	{
		return MAC.getMacList();
	}

	/**
	 * Returns a <code>ConnectionInfo</code> object containing the details of
	 * the connection. Can be called as soon as the connection has been
	 * established (successfully connected).
	 * 
	 * @return A <code>ConnectionInfo</code> object.
	 * @throws IOException
	 *             In case of any failure behind the scenes.
	 */
	public synchronized ConnectionInfo getConnectionInfo() throws IOException
	{
		if (tm == null)
			throw new IllegalStateException(
					"Cannot get details of connection, you need to establish a connection first.");
		return tm.getConnectionInfo(1);
	}

	/**
	 * Open a new session on this connection. Works only after one has passed
	 * successfully the authentication step. There is no limit on the number of
	 * concurrent sessions.
	 * 
	 * @return A <code>Session</code> object.
	 * @throws IOException
	 */
	public synchronized Session openSession() throws IOException
	{
		if (tm == null)
			throw new IllegalStateException("Cannot open session, you need to establish a connection first.");

		if (!authenticated)
			throw new IllegalStateException("Cannot open session, connection is not authenticated.");

		return new Session(cm);
	}

	// /**
	// * Open a new SFTP client session on this connection. Works only after one
	// * has passed successfully the authentication step. One can open as many
	// * concurrent sessions as one wants.
	// *
	// * @return A <code>Session</code> object.
	// * @throws IOException
	// */
	// public synchronized SFTPClient openSFTP() throws IOException
	// {
	// if (tm == null)
	// throw new IllegalStateException("Cannot open sftp session, you need to
	// establish a connection first.");
	//
	// if (!authenticated)
	// throw new IllegalStateException("Cannot open sftp session, connection is
	// not authenticated.");
	//
	// return new SFTPClient(cm);
	// }

	/**
	 * Create a very basic <code>SCPClient</code> that can be used to copy
	 * files from/to the SSH2 server. <br>
	 * Works only after one has passed successfully the authentication step.
	 * There is no limit on the number of concurrent SCP clients.
	 * 
	 * @return A <code>SCPClient</code> object.
	 * @throws IOException
	 */
	public synchronized SCPClient createSCPClient() throws IOException
	{
		if (tm == null)
			throw new IllegalStateException("Cannot create SCP client, you need to establish a connection first.");

		if (!authenticated)
			throw new IllegalStateException("Cannot create SCP client, connection is not authenticated.");

		return new SCPClient(this);
	}

	/**
	 * Creates a new <code>LocalPortForwarder</code>. A
	 * <code>LocalPortForwarder</code> forwards TCP/IP connections to a local
	 * port via the secure tunnel to another host (which may or may not be
	 * identical to the remote SSH2 server). <br>
	 * Works only after one has passed successfully the authentication step.
	 * There is no limit on the number of concurrent forwardings.
	 * 
	 * @return A <code>LocalPortForwarder</code> object.
	 * @throws IOException
	 */
	public synchronized LocalPortForwarder createLocalPortForwarder(int local_port, String host_to_connect,
			int port_to_connect) throws IOException
	{
		if (tm == null)
			throw new IllegalStateException("Cannot forward ports, you need to establish a connection first.");

		if (!authenticated)
			throw new IllegalStateException("Cannot forward ports, connection is not authenticated.");

		return new LocalPortForwarder(cm, local_port, host_to_connect, port_to_connect);
	}

	/**
	 * Creates a new <code>LocalStreamForwarder</code>. A
	 * <code>LocalStreamForwarder</code> forwards an Input- and Outputstream
	 * pair via the secure tunnel to another host (which may or may not be
	 * identical to the remote SSH2 server).
	 * 
	 * @return A <code>LocalStreamForwarder</code> object.
	 * @throws IOException
	 */
	public synchronized LocalStreamForwarder createLocalStreamForwarder(String host_to_connect, int port_to_connect)
			throws IOException
	{
		if (tm == null)
			throw new IllegalStateException("Cannot forward, you need to establish a connection first.");

		if (!authenticated)
			throw new IllegalStateException("Cannot forward, connection is not authenticated.");

		return new LocalStreamForwarder(cm, host_to_connect, port_to_connect);
	}

	/**
	 * Sets the parameters for the diffie-hellman group exchange. Unless you
	 * know what you are doing, you will never need this. Default values are
	 * defined in the <code>DHGexParameters</code> class.
	 * 
	 * @param dgp
	 *            non null.
	 */
	public synchronized void setDHGexParameters(DHGexParameters dgp)
	{
		if (dgp == null)
			throw new IllegalArgumentException();

		dhgexpara = dgp;
	}

	/**
	 * Unless you know what you are doing, you will never need this.
	 * 
	 * @param ciphers
	 */
	public synchronized void setClient2ServerCiphers(String[] ciphers)
	{
		if (ciphers == null)
			throw new IllegalArgumentException();
		BlockCipherFactory.checkCipherList(ciphers);
		cryptoWishList.c2s_enc_algos = ciphers;
	}

	/**
	 * Unless you know what you are doing, you will never need this.
	 * 
	 * @param macs
	 */
	public synchronized void setClient2ServerMACs(String[] macs)
	{
		if (macs == null)
			throw new IllegalArgumentException();
		MAC.checkMacList(macs);
		cryptoWishList.c2s_mac_algos = macs;
	}

	/**
	 * Unless you know what you are doing, you will never need this.
	 * 
	 * @param ciphers
	 */
	public synchronized void setServer2ClientCiphers(String[] ciphers)
	{
		if (ciphers == null)
			throw new IllegalArgumentException();
		BlockCipherFactory.checkCipherList(ciphers);
		cryptoWishList.s2c_enc_algos = ciphers;
	}

	/**
	 * Unless you know what you are doing, you will never need this.
	 * 
	 * @param macs
	 */
	public synchronized void setServer2ClientMACs(String[] macs)
	{
		if (macs == null)
			throw new IllegalArgumentException();
		MAC.checkMacList(macs);
		cryptoWishList.s2c_mac_algos = macs;
	}
}
