
package ch.ethz.ssh2.packets;

import java.io.IOException;

import java.math.BigInteger;

/**
 * TypesReader.
 * 
 * @author Christian Plattner, plattner@inf.ethz.ch
 * @version $Id: TypesReader.java,v 1.1 2005/05/26 14:53:26 cplattne Exp $
 */
public class TypesReader
{
	byte[] arr;
	int pos = 0;
	int max = 0;

	public TypesReader(byte[] arr)
	{
		this.arr = arr;
		pos = 0;
		max = arr.length;
	}

	public TypesReader(byte[] arr, int off)
	{
		this.arr = arr;
		this.pos = off;
		this.max = arr.length;

		if ((pos < 0) || (pos > arr.length))
			throw new IllegalArgumentException("Illegal offset.");
	}

	public TypesReader(byte[] arr, int off, int len)
	{
		this.arr = arr;
		this.pos = off;
		this.max = off + len;

		if ((pos < 0) || (pos > arr.length))
			throw new IllegalArgumentException("Illegal offset.");

		if ((max < 0) || (max > arr.length))
			throw new IllegalArgumentException("Illegal length.");
	}

	public int readByte() throws IOException
	{
		if (pos >= max)
			throw new IOException("Packet too short.");

		return (arr[pos++] & 0xff);
	}

	public byte[] readBytes(int len) throws IOException
	{
		if ((pos + len) > max)
			throw new IOException("Packet too short.");

		byte[] res = new byte[len];

		System.arraycopy(arr, pos, res, 0, len);
		pos += len;

		return res;
	}

	public void readBytes(byte[] dst, int off, int len) throws IOException
	{
		if ((pos + len) > max)
			throw new IOException("Packet too short.");

		System.arraycopy(arr, pos, dst, off, len);
		pos += len;
	}

	public boolean readBoolean() throws IOException
	{
		if (pos >= max)
			throw new IOException("Packet too short.");

		return (arr[pos++] != 0);
	}

	public int readUINT32() throws IOException
	{
		if ((pos + 4) > max)
			throw new IOException("Packet too short.");

		return ((arr[pos++] & 0xff) << 24) | ((arr[pos++] & 0xff) << 16) | ((arr[pos++] & 0xff) << 8)
				| (arr[pos++] & 0xff);
	}

	public long readUINT64() throws IOException
	{
		if ((pos + 8) > max)
			throw new IOException("Packet too short.");

		return ((arr[pos++] & 0xff) << 56) | ((arr[pos++] & 0xff) << 48) | ((arr[pos++] & 0xff) << 40)
				| ((arr[pos++] & 0xff) << 32) | ((arr[pos++] & 0xff) << 24) | ((arr[pos++] & 0xff) << 16)
				| ((arr[pos++] & 0xff) << 8) | (arr[pos++] & 0xff);
	}

	public BigInteger readMPINT() throws IOException
	{
		byte raw[] = readByteString();

		BigInteger b = new BigInteger(raw);

		return b;
	}

	public byte[] readByteString() throws IOException
	{
		int len = readUINT32();

		if ((len + pos) > max)
			throw new IOException("Malformed SSH string.");

		byte[] res = new byte[len];
		System.arraycopy(arr, pos, res, 0, len);
		pos += len;
		return res;
	}

	public String readString(String charSetName) throws IOException
	{
		int len = readUINT32();

		if ((len + pos) > max)
			throw new IOException("Malformed SSH string.");

		String res = new String(arr, pos, len, charSetName);
		pos += len;

		return res;
	}

	public String readString() throws IOException
	{
		int len = readUINT32();

		if ((len + pos) > max)
			throw new IOException("Malformed SSH string.");

		String res = new String(arr, pos, len);
		pos += len;

		return res;
	}

	public String[] readNameList() throws IOException
	{
		String enclist = readString();

		/* StringTokenizer is not available in J2ME */

		int numtoken = 1;

		for (int i = 0; i < enclist.length(); i++)
		{
			if (enclist.charAt(i) == ',')
				numtoken++;
		}

		String list[] = new String[numtoken];
		int nextfield = 0;

		for (int i = 0; i < numtoken; i++)
		{
			if (nextfield >= enclist.length())
			{
				list[i] = "";
			}
			else
			{
				int idx = enclist.indexOf(',', nextfield);
				if (idx == -1)
					idx = enclist.length();
				list[i] = enclist.substring(nextfield, idx);
				nextfield = idx + 1;
			}
		}

		return list;
	}

	public int remain()
	{
		return max - pos;
	}

}
